#! /usr/local/bin/gawk -f -
### ====================================================================
### Implement a simple manual-page-to-TeXinfo filter, to provide info
### access to Unix manual pages.
###
### Usage:
###	awk [-v Filename=xxx] -f man2texi.awk infile > outfile
###	... | awk [-v Filename=xxx] -f man2texi.awk | ...
###
### [19-Dec-2002]
### ====================================================================

BEGIN						{ initialize() }

### (FNR % 5 == 0)				{ comment("# " FNR " \"" get_filename() "\"") }

/^[.]\\" *-[*]-.roff-[*]-/ 			{ comment("-*-texinfo-*-"); next }

/^[.]\\"/ 					{ comment(substr($0,4)); next }

/^[.]/						{ comment("TROFF INPUT: " $0) } # for debugging only

/^[.]PD/					{ comment("TROFF INPUT: " $0); next }

/^[.]ds/					{ define_string($0); next }

(ItemPending[Level] != 0)			{ listitem() }

						{ $0 = preprocess($0) }

/^[.][BIR] /					{ font($0); next }

/^[.][BIR][BIR] /				{ font_pair($0); next }

/^[.]/						{ troff_command($0); next }

						{ copy($0); next }

END						{ terminate() }

### ====================================================================

function add_menu(section,name)
{
    Menu[section] = Menu[section] "* " node_name(name) "::\n"
}


function begin_environment(s)
{
    ++Level
    EnvName[Level] = s
    ItemPending[Level] = 0
    ItemCount[Level] = 0
    writeln("")
    hline()
    if (s == "table")
	writeln("@" s " @code")
    else
	writeln("@" s)
    ## writeln("@comment DEBUG: Level = " Level)
    ## dump_environment()
}


function comment(s)
{
    writeln("@comment " s)
}


function copy(s)
{
    writeln(s)
}


function current_environment()
{
    return ((Level <= 0) ? "" : EnvName[Level])
}


function define_string(s, t)
{
    ## Typical value: .ds PX \s-1POSIX\s+1
    t = s
    sub("[.]ds +","",t)
    if (match(t,"^[A-Za-z][A-Za-z]+"))
	StringValue[substr(t,RSTART,RLENGTH)] = trim(substr(t,RSTART+RLENGTH))
    else
	warning("unrecognized define-string command: [" s "]")
}


function dump_environment( k,t)
{
    if (Level > 0)
    {
	t = "@comment DEBUG: dump_environment(): "
	for (k = 1; k <= Level; ++k)
	    t = sprintf("%s\t%d:%s", t, k, EnvName[k])
	writeln(t)
    }
    else
	writeln("@comment DEBUG: dump_environment(): empty stack")
}


function end_environment(expect)
{
    ## dump_environment()
    if (Level <= 0)
    {
	warning("mismatched environments caused stack underflow: expected [" expect "]")
	Level = 0
    }
    else
    {
	writeln("")
	writeln("@end " EnvName[Level])
	writeln("")
	hline()
	if (EnvName[Level] != expect)
	    warning("bad environment nesting: found end of " \
		    EnvName[Level] " but expected end of " expect)
	EnvName[Level] = "UNSET"
	ItemPending[Level] = 0
	ItemCount[Level] = 0
	--Level
    }
}


function end_open_environments()
{
    ## writeln("@comment DEBUG: end_open_environments()")
    ## dump_environment()

    while (current_environment() != "")
	end_environment(current_environment())
}


function font(s)
{
    writeln("@" font_code(substr(s,2,1)) "{" unquote(substr(s,3)) "}")
}


function font_code(s, fontname)
{
    if (s == "B")
	fontname = "b"
    else if (s == "I")
	fontname = "i"
    else if (s == "R")
	fontname = "r"
    else
	fontname = "unknownfont"
    return (fontname)
}


function font_pair(s, args,k,n,t,t1,t2)
{
    t1 = font_code(substr(s,2,1))
    t2 = font_code(substr(s,3,1))
    n = troff_args(s,args)
    t = ""
    for (k = 1; k <= n; )
    {
	t = sprintf("%s@%s{%s}", t, t1 ,args[k++])
	if (k <= n)
	    t = sprintf("%s@%s{%s}", t, t2 ,args[k++])
    }
    writeln(t)
}


function font_inline(s, fontchar,r)
{
    ## Originally we just assumed well-balance font changes, with
    ## each \fX closed by \fP.  However, many manual pages have sloppy
    ## font changes with unbalanced switching, or with \fR used where
    ## \fP should have been used.  Handle them, but warn about them.

    ## gsub("\\\\fB", "@b{", s)
    ## gsub("\\\\fC", "@t{", s)
    ## gsub("\\\\fR", "@r{", s)
    ## gsub("\\\\fI", "@i{", s)
    ## gsub("\\\\fP", "}", s)

    while (match(s,"\\\\f[BCRIP]"))
    {
	fontchar = substr(s,RSTART+RLENGTH-1,1)
	if (fontchar == "P")	# end previous font
	{
	    if (Last_Font == "")
	    {
		r = ""
		warning("unbalanced \\fP ignored")
	    }
	    else
		r = "}"
	    Last_Font = ""
	}
	else			# start new font
	{
	    if (Last_Font == "")	# have correct font nesting
		r = ("@" Fontmap[fontchar] "{")
	    else
	    {
		warning("new font " fontchar \
			" started without closing previous font " Last_Font)
		if (fontchar == "R")	# default font: treat like \fP
		{
		    r = "}"
		    fontchar = ""
		}
		else
		    r = ("}@" Fontmap[fontchar] "{")
	    }
	    Last_Font = fontchar
	}
	s = (substr(s,1,RSTART-1) r substr(s,RSTART+RLENGTH))
    }
    return (s)
}


function get_filename()
{
    return ((Filename == "") ? FILENAME : Filename)
}

function hline()
{
    writeln("@c ---------------------------------------------------------------------")
}


function initialize( s)
{
    "date" | getline Current_Date_and_Time
    close("date")

    Fontmap["B"] = "b"
    Fontmap["C"] = "t"
    Fontmap["I"] = "i"
    Fontmap["R"] = "r"

    Level = 0

    ItemPending[0] = 0
    ItemPending[1] = 0

    StringValue["lq"] = "``"		# complex definition in gawk.man
    StringValue["rq"] = "''"		# complex definition in gawk.man

    Nlines = 1

    writeln("\\input texinfo   @c -*-texinfo-*-")
    s = get_filename()
    sub("[.][^.]*$","",s)		# remove any extension
    sub("^.*/","",s)			# remove any leading path component

    writeln("@setfilename " s ".info")
    writeln("@comment @documentencoding ISO-8859-1")
    writeln("@comment ===============================================================")
    writeln("@comment WARNING: Do NOT edit this file.  It was produced automatically")
    writeln("@comment by man2info on " Current_Date_and_Time)
    writeln("@comment for " ENVIRON["USER"] "@" ENVIRON["HOST"])
    writeln("@comment from " get_filename())
    writeln("@comment in " ENVIRON["PWD"])
    writeln("@comment ===============================================================")
}


function listitem()
{
    if (ItemCount[Level] == 0)
    {
	ItemPending[Level] = 0
	begin_environment("table")
    }
    else
	writeln("")
    ## writeln("@comment DEBUG: ItemCount[" Level "] = " ItemCount[Level]+1)
    write("@item ")			# NB: no newline here!
    ItemCount[Level]++
    ItemPending[Level] = 0
}


function node_name(s)
{
    gsub(" *[,] *", " ", s)		# no commas allowed in node text
    gsub(" *[:] *", " .. ", s)		# no colons allowed in menu text
    gsub("[ ]+", " ", s)		# squeeze space
    return (s)
}


function preprocess(s)
{
    ## Expand string definitions first
    while (match(s,"\\\\[*][(][A-Za-z][A-Za-z]"))
	s = substr(s,1,RSTART-1) \
		string_value(substr(s,RSTART+3,RLENGTH-3)) \
		substr(s,RSTART+RLENGTH)

    gsub("\\201","",s)			# remove emacs-inserted prefix on binary (accented) characters

    gsub("@", "@@", s)
    gsub("{", "@{", s)
    gsub("}", "@}", s)
    gsub("_", "@t{_}", s)
    gsub("[.]\\\\[|][.]\\\\[|][.]\\\\[.]", "@enddots{}", s)
    gsub("[.]\\\\[|][.]\\\\[|][.]", "@dots{}", s)
    gsub("\\\\\\^", "", s)
    gsub("\\\\[&]", "", s)
    gsub("\\\\-", "@minus{}", s)
    gsub("\\\\ ", "@tie{}", s)
    gsub("\\\\[(]bu", "@bullet{}", s)
    gsub("\\\\[(]em", "@r{---}", s)
    gsub("\\\\[(]mi", "@r{--}", s)

    ## This does not work when used inside an example:
    ## gsub("\\\\[(]->", "@math{\\rightarrow}", s)
    ## so use a verbatim arrow:
    gsub("\\\\[(]->", "@t{-->}", s)

    gsub("\\\\e\"", "@t{\\\"}", s)
    gsub("\\\\e", "@t{\\}", s)
    gsub("\\\\[|]", "@dmn{}", s)

    ## Handle font size change
    while (match(s,"\\\\s-1[^\\\\]*\\\\s[+]1"))
	s = substr(s,1,RSTART-1) smallcaps(substr(s,RSTART+4,RLENGTH-8)) \
		substr(s,RSTART+RLENGTH)

    ## Handle another style of font size change
    while (match(s,"\\\\s-1[^\\\\]*\\\\s0"))
	s = substr(s,1,RSTART-1) smallcaps(substr(s,RSTART+4,RLENGTH-7)) \
		substr(s,RSTART+RLENGTH)

    ## Remove any remaining font size changes
    gsub("\\\\s[-+]?[0-9]","",s)

    s = font_inline(s)
    return (s)
}


function print_menu(section)
{
    print "@c DEBUG: print_menu(\"" section "\")"
    if (section in Menu)
    {
	print ""
	print "@ifnottex"
	print "@menu"
	print Menu[section]
	print "@end menu"
	print "@end ifnottex"
	print ""
    }
}


function smallcaps(s)
{
    ## TeXinfo documentation says that all-caps arguments
    ## of a smallcaps font should be in lowercase.
    ## print "DEBUG: smallcaps(" s ")" > "/dev/stderr"
    return ("@sc{" ((s ~ "^[A-Z]+$") ? tolower(s) : s) "}")
}


function string_value(s)
{
    return ((s in StringValue) ? StringValue[s] : "??unknown-string??")
}


function troff_command(s, t)
{
    if (s ~ "^[.]IP")
	writeln("")
    else if (s ~ "^[.]PP")
    {
	## NO: end_open_environments()
	writeln("")
    }
    else if (s ~ "^[.]PD")		# used in gawk.man
    {
	## NO: end_open_environments()
	writeln("")
    }
    else if (s ~ "^[.]RE")
    {
	if (current_environment() == "table")	# implicit end of open table
	    end_environment("table")
	end_environment("quotation")
    }
    else if (s ~ "^[.]RS")
	begin_environment("quotation")
    else if (s ~ "^[.]SH")
    {
	end_open_environments()
	t = unquote(substr(s,4))
	add_menu("Top",t)
	write_node(t)
	writeln("@section " t)
	writeln("")
	Section = t
    }
    else if (s ~ "^[.]SS")
    {
	end_open_environments()
	t = unquote(substr(s,4))
	add_menu(Section,t)
	write_node(t)
	writeln("@subsection " t)
	writeln("")
    }
    else if (s ~ "^[.]TH")
    {
	end_open_environments()
	write_node("Top")
	writeln("@chapter " unquote(substr(s,4)))
	writeln("")
    }
    else if (s ~ "^[.]TP")
    {
	## Because the item label is not known until the next
	## non-comment line, we cannot output the item translation
	## yet, but we have to record that an item is pending.
	ItemPending[Level] = 1
    }
    else if (s ~ "^[.]nf")
	begin_environment("example")
    else if (s ~ "^[.]fi")
	end_environment("example")
    else
	writeln("@comment " s)
}


function terminate( k)
{
    end_open_environments()
    writeln("@bye")
    for (k = 1; k < Nlines; ++k)
    {
	printf("%s", Lines[k])
	if (Lines[k] ~ "@chapter")
	    print_menu("Top")
	else if (Lines[k] ~ "@section")
	    print_menu(trim_newline(substr(Lines[k],9)))
    }
}


function trim(s)
{
    gsub(/^[ \t]+/,"",s)
    gsub(/[ \t]+$/,"",s)
    return (s)
}


function trim_newline(s)
{
    gsub(/[\n]/," ",s)
    return (trim(s))
}


function troff_args(s,args, kk,n)
{
    ## Split a troff command into args[] and return the count

    ## print "DEBUG: troff_args(" s ")"

    split("",args,"")			# clear the array
    s = trim(s)
    n = index(s," ")			# find start of args (after initial .CMD)
    if (n == 0)
    {
	warning("expected one or more arguments in [" s "]")
	args[1] = s
	return (1)
    }

    s = substr(s,n+1)			# discard initial ".CMD "

    n = 0
    while (length(s) > 0)
    {
	## print "DEBUG: troff_args(): s = [" s "]"
	s = trim(s)			# assertion: no leading/trailing space
	if (match(s,"^\"[^\"]*\""))	# quoted argument
	{
	    args[++n] = substr(s,RSTART+1,RLENGTH-2)
	    s = substr(s,RSTART+RLENGTH)
	}
	else if (match(s,"^[^ ]+"))	# argument is ordinary word
	{
	    args[++n] = substr(s,RSTART,RLENGTH)
	    s = substr(s,RSTART+RLENGTH)
	}
	else				# should never happen
	{
	    args[++n] = s
	    s = ""
	}
	s = trim(s)		 	# preserve assertion: no leading/trailing space
    }

    ## for (kk = 1; kk <= n; ++kk) print "DEBUG: troff_args(): args[" kk "] = [" args[kk] "]"

    return (n)
}


function unquote(s)
{
    ## strip surrounding quotes and whitespace
    gsub("^ \"","",s)
    gsub("\" *$","",s)
    return (trim(s))
}


function warning(message)
{
    print get_filename() ":" FNR ":%%" message >"/dev/stderr"

    ## Ensure that the warning is preserved in the output as a comment
    ## (unless we are typesetting a verbatim example)!
    if (current_environment() != "example")
	writeln("\n@comment WARNING: " get_filename() ":" FNR ":%%" message "\n")
}


function write(s)
{
    Lines[Nlines++] = s
}


function write_node(s)
{
    writeln("")
    writeln("@node " node_name(s))
    writeln("")
}


function writeln(s)
{
    sub(" *$","",s)			# trim trailing space
    Lines[Nlines++] = (s "\n")
}
