#! /bin/sh
### ====================================================================
### Implement a simple manual-page-to-TeXinfo filter, to provide info
### access to Unix manual pages.
###
### Usage:
###
###	man2texi [ -author ] [ -copyright ] [ -filename filename ] \
###		[ -help ] [ -version ] infile > outfile
###
###	man2texi [ -author ] ( -batch | -dvi | -info ) [ -copyright ] \
###		 [ -filename filename ] [ -help ] [ -version ] infile(s)
###
###	... | man2texi -filename filename | ...
###
### The -filename filename value must be set when man2texi is used in
### a pipe if the output file is to be a correct TeXinfo document.
### [19-Dec-2002]
### ====================================================================

AWK=/usr/local/bin/gawk
MAKEINFO=/usr/local/bin/makeinfo
MV=/usr/local/bin/mv
SED=/usr/local/bin/sed
SHRLIBDIR=.
TEX="/usr/local/bin/tex"
TRACE=yes

author()
{
    ## Handle the -author option.

    cat 1>&2 <<EOF

Author:
	Nelson H. F. Beebe
	Center for Scientific Computing
	University of Utah
	Department of Mathematics, 110 LCB
	155 S 1400 E RM 233
	Salt Lake City, UT 84112-0090
	USA
	Email: beebe@math.utah.edu, beebe@acm.org,
	       beebe@computer.org, beebe@ieee.org (Internet)
	WWW URL: http://www.math.utah.edu/~beebe
	Telephone: +1 801 581 5254
	FAX: +1 801 581 4148

EOF
}

copyright()
{
    ## Handle the -copyright option.

    cat 1>&2 <<EOF

########################################################################
########################################################################
########################################################################
###                                                                  ###
###  man2texi: convert UNIX manual pages to TeXinfo format           ###
###                                                                  ###
###              Copyright (C) 2002 Nelson H. F. Beebe               ###
###                                                                  ###
### This program is covered by the GNU General Public License (GPL), ###
### version 2 or later, available as the file COPYING in the program ###
### source distribution, and on the Internet at                      ###
###                                                                  ###
###               ftp://ftp.gnu.org/gnu/GPL                          ###
###                                                                  ###
###               http://www.gnu.org/copyleft/gpl.html               ###
###                                                                  ###
### This program is free software; you can redistribute it and/or    ###
### modify it under the terms of the GNU General Public License as   ###
### published by the Free Software Foundation; either version 2 of   ###
### the License, or (at your option) any later version.              ###
###                                                                  ###
### This program is distributed in the hope that it will be useful,  ###
### but WITHOUT ANY WARRANTY; without even the implied warranty of   ###
### MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    ###
### GNU General Public License for more details.                     ###
###                                                                  ###
### You should have received a copy of the GNU General Public        ###
### License along with this program; if not, write to the Free       ###
### Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,   ###
### MA 02111-1307 USA                                                ###
########################################################################
########################################################################
########################################################################

EOF
}

help()
{
    ## Handle the -help option

    cat 1>&2 <<EOF

Usage:
	man2texi [ -? ] [ -author ] [ -copyright ] [ -filename filename ]
		 [ -help ] [ -version ] infile(s) [> outfile]

	man2texi [ -author ] ( -batch | -dvi | -info ) [ -copyright ]
		 [ -filename filename ] [ -help ] [ -version ] infile(s)

	... | man2texi -filename filename | ...

Command-line options may begin with a double hyphen, and may be
abbreviated to any unique prefix.

EOF

}

overwrite()
{
    cat 1>&2 <<EOF

ERROR: you have asked me to overwrite the input file $texifile, which
is probably not what you meant! man2texi expects UNIX manual page
input files, NOT TeXinfo files

EOF
}

version()
{
    ## Handle the -version option

    cat 1>&2 <<EOF

This is man2texi version 0.00 of 19-Dec-2002

EOF
}


### Option defaults:
BATCH=no
DVI=no
INFO=no

### File lists
FILENAME=
FILES=

while test "$#" -gt 0
do
	case "$1" in
		-author | -autho | -auth | -aut | -au | -a | \
		--author | --autho | --auth | --aut | --au | --a )
			author
			exit 0
			;;

		-batch | -batc | -bat | -ba | -b | \
		--batch | --batc | --bat | --ba | --b )
			BATCH=yes
			;;

		-copyright | -copyrigh | -copyrig | -copyri | -copyr | \
		-copy | -cop | -co | -c | \
		--copyright | --copyrigh | --copyrig | --copyri | --copyr | \
		--copy | --cop | --co | --c )
			copyright
			exit 0
			;;

		-dvi | -dv | -d | \
		--dvi | --dv | --d )
			BATCH=yes
			DVI=yes
			;;

		-filename | -filenam | -filena | -filen | -file | -fil | -fi | -f | \
		--filename | --filenam | --filena | --filen | --file | --fil | --fi | \
		--f )
			shift
			FILENAME="$1"
			;;

		-info | -inf | -in | -i | \
		--info | --inf | --in | --i )
			BATCH=yes
			INFO=yes
			;;

		-[?] | -help | -hel | -he | -h | \
		--[?] | --help | --hel | --he | --h )
			help
			exit 0
			;;

		-version | -versio | -versi | -vers | -ver | -ve | -v | \
		--version | --versio | --versi | --vers | --ver | --ve | --v )
			version
			exit 0
			;;

		-*)
			echo Unrecognized option: $1
			help
			exit 1
			;;
		*)
			test -z "$FILENAME" && FILENAME="$1"
			FILES="$FILES $1"
			;;

	esac
	shift
done

test -z "$FILENAME" && FILENAME=CHANGE-THIS-FILENAME

### echo "DEBUG: FILES = $FILES"
### echo "DEBUG: FILENAME = $FILENAME"

if test "$BATCH" = "yes"
then
	for infile in $FILES
	do
		b=`echo $infile | sed -e 's/[.][^.]*$//' `
		dvifile=$b.dvi
		infofile=$b.info
		texifile=$b.texi
		texlogfile=$b.tlg

		test "$TRACE" = yes && echo $infile " -> " $texifile

		if test "$infile" = "$texifile"
		then
			overwrite
			help
			exit 1
		else
			test -f $texifile && $MV -f $texifile $texifile.org
			$AWK -v Filename="$infile" -f $SHRLIBDIR/man2texi.awk $infile |
				$SED -f $SHRLIBDIR/iso8859-1-to-TeXinfo.sed > $texifile
			if test "$DVI" = "yes"
			then
				test "$TRACE" = yes && echo $texifile " -> " $dvifile
				## Must run TeX twice to get cross-references correct
				$TEX $texifile < /dev/null > $texlogfile
				$TEX $texifile < /dev/null >> $texlogfile
			fi
			if test "$INFO" = "yes"
			then
				test "$TRACE" = yes && echo $texifile " -> " $infofile
				$MAKEINFO --force --no-split $texifile
			fi
		fi
	done
else # not batch: output to stdout (and thus no -dvi or -info support!)
	$AWK -v Filename="$FILENAME" -f $SHRLIBDIR/man2texi.awk $FILES |
		$SED -f $SHRLIBDIR/iso8859-1-to-TeXinfo.sed
fi
